/* $Id: path.c,v 2.0 1998/03/25 05:43:54 svmcguir Exp svmcguir $ */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "path.h"
#include "util.h"
#include "ranlib.h"

void pathDataAllocate(pathData *pd){
  /* Allocates memory for arrays in pathData structure based on the values
   * of pd->underliers, pd->times and pd->paths.  These obviously must be
   * set before calling the function.
   */

  allocate2dDouble( &(pd->covariance), pd->underliers, pd->underliers);
  allocate1dDouble( &(pd->growthRate), pd->underliers);
  allocate1dDouble( &(pd->dividendRate), pd->underliers);
  allocate1dDouble(  &(pd->cov1), pd->underliers*pd->underliers);
  allocate1dDouble(  &(pd->fgrowth), pd->underliers);
  allocate1dDouble(  &(pd->parm), (pd->underliers*(pd->underliers+3)/2 +1));
  allocate3dDouble( &(pd->price), pd->times,pd->meshPaths,pd->underliers);
}

void pathDataDeallocate(pathData *pd){
  /* Deallocates memory for arrays in pathData structure based on the value
   * of pd->underliers, pd->times and pd->paths.
   */

  deallocate2dDouble( &(pd->covariance), pd->underliers, pd->underliers);
  deallocate1dDouble( &(pd->growthRate), pd->underliers);
  deallocate1dDouble( &(pd->dividendRate), pd->underliers);
  deallocate1dDouble(  &(pd->cov1), pd->underliers*pd->underliers);
  deallocate1dDouble(  &(pd->fgrowth), pd->underliers);
  deallocate1dDouble(  &(pd->parm), (pd->underliers*(pd->underliers+3)/2 +1));
  deallocate3dDouble( &(pd->price), pd->times,pd->meshPaths,pd->underliers);
}

void pathParamsInit(pathData *pd){
  /* Uses pd->underliers, pd->covariance[i][j], pd->growthRate[i], and
   * pd-phrase to initialize the private data and the RANLIB library.
   *
   * After initialization, whenever we want numbers from the
   * distribution described in pd we call
   * genmn(pd->parm,vector,pd->cov1) where vector is a double array of
   * dimension [0:pd->underliers-1].  The random numbers go in vector.
   */

  int i,j,k;

  pd->constant = pow(2*M_PI,pd->underliers);
  for(i=0; i<pd->underliers; i++){
    pd->constant *= pd->covariance[i][i];
  }
  pd->constant = 1.0/sqrt(pd->constant);

  /* Allocate growth, same dimension as prices */
  allocate1dDouble(&(pd->growth),pd->underliers);


  /* Copy the 2d covariance matrix into the 1d scratch matrix */
  k = 0;
  for(i=0;i<pd->underliers;i++){
    for(j=0;j<pd->underliers;j++){
      pd->cov1[k] = pd->covariance[i][j];
      k++;
    }
  }

  /* Copy double growthRate array into double fgrowth array */
  /* Subtractthe dividend rate from the growth rate */
  for(i=0;i<pd->underliers;i++){
    pd->fgrowth[i] = pd->growthRate[i] - pd->dividendRate[i];
  }

  /* Give the data to RANLIB and let it do what it wants in parm */
  setgmn(pd->fgrowth,pd->cov1,pd->underliers,pd->parm);

  /* Hash the phrase into RNG seeds */
  phrtsd(pd->phrase,&(pd->seed1),&(pd->seed2));

  /* Initialize the RNG with the seeds */
  setall(pd->seed1,pd->seed2);
}


void pathDataFill(pathData *pd){
  /* Uses pd->params, pd->paths, and pd->times to generate paths which are
   * placed in pd->price[p][t][u].  Make sure all these are initialized and
   * allocated first.  Each path starts with the value in pd->price[0][p][u]
   * so set these also.
   *
   * pd->params describes the random numbers we get by calling
   * genmn(pd->params.parm, randomNum, pd->params.cov1).  To get the value in
   * the next time slice, we grow the values in the previous time slice by the
   * random numbers.  Each underlier gets its own growth rate.
   */


  int time, path, underlier,basePath, randomPath;
  double *randomNum, std;

  randomNum = (double*)calloc(pd->underliers,sizeof(double));

  for(time=1; time<pd->times; time++){
    for(path=0;path<pd->meshPaths;path++){
      if(path % pd->clusterSize == 0){
        genmn(pd->parm, randomNum, pd->cov1);
        for(underlier = 0; underlier < pd->underliers; underlier++){
	  randomPath = (int)(pd->meshPaths*drand48());
          pd->price[time][path][underlier] = 
            exp(randomNum[underlier]-pd->covariance[underlier][underlier]/2)*
            pd->price[time-1][path][underlier];
        }
        basePath = path;
      }else{
        genmn(pd->parm, randomNum, pd->cov1);
        for(underlier = 0; underlier < pd->underliers; underlier++){
          pd->price[time][path][underlier] =
            exp(randomNum[underlier]-pd->covariance[underlier][underlier]/2)*
            pd->price[time-1][basePath][underlier];
        }
      }
    }
  }

  free(randomNum);
}


double pathTransProb(pathData *pd, double *price1, double *price2){
  /* Returns the probability of a transition from price vector price1[]
   * at one timeslice to price vector price2[] at the next timeslice.
   */

  int i;
  double exponential;

  for(i=0; i<pd->underliers; i++){
    pd->growth[i] = log(price2[i]/price1[i]) +
      pd->covariance[i][i]/2;
  }   

  /* Calculate the exponential term */
  exponential = 0;
  for(i=0; i<pd->underliers; i++){
    exponential += (pd->growth[i] - pd->fgrowth[i])*(pd->growth[i] -
                    pd->fgrowth[i])/pd->covariance[i][i];
  }
  exponential = exp(-0.5*exponential);

  /* Put them together and we have our result */
  return(pd->constant*exponential);
}        

void pathDataOut(pathData *pd,double t){ 
  int time, path, underlier,basePath; 
  double *x,*y; 
  int n,l; 
  int ierr; 
  n = pd->underliers*pd->times*pd->meshPaths; 
  x = (double*)calloc(n,sizeof(double)); 
  y = (double*)calloc(n,sizeof(double)); 
  l = 0; 

  for(underlier = 0; underlier < pd->underliers; underlier++){ 
    for(path=0;path<pd->meshPaths;path++){ 
      for(time=0; time<pd->times; time++){ 
        x[l]= (double)time; 
        y[l] = pd->price[time][path][underlier]; 
        /* printf("%i %i %lf %lf \n",path,time,x[l],y[l]);  */
        l++; 
      } 
    } 
  } 
  ierr = java_ser("paths",t,x,y,l); 

  free(x); 
  free(y); 
} 


void pathPair(pathData *pd, double **path, double **antiPath){
  /* Generates a random path and its antithetic path.
   *
   * When finished, path[time][underlier] contains the path
   * and antiPath[time][underlier] contains the antithetic path.
   */

  int time, underlier;
  double  delta, antiGrowth;
  double *randomNum;

  allocate1dDouble(&randomNum,  pd->underliers);  

  for(underlier=0; underlier<pd->underliers; underlier++){
    path[0][underlier] = pd->price[0][0][underlier];
    antiPath[0][underlier] = pd->price[0][0][underlier];
  }
  
  for(time=1; time < pd->times; time++){
    genmn(pd->parm, randomNum, pd->cov1);
    for(underlier=0; underlier<pd->underliers; underlier++){
      path[time][underlier] =
        exp(randomNum[underlier]-pd->covariance[underlier][underlier]/2)*
        path[time-1][underlier];

      delta = randomNum[underlier] - pd->fgrowth[underlier]; 
      antiGrowth = pd->fgrowth[underlier] - delta; 

      antiPath[time][underlier] =
        exp(antiGrowth-pd->covariance[underlier][underlier]/2)*
        antiPath[time-1][underlier];
    }
  }
  deallocate1dDouble(&randomNum,  pd->underliers); 
}
